import os
import sys

from loguru import logger

from alist_mikananirss.alist import Alist, AlistConfig
from alist_mikananirss.bot import NotificationBot, PushPlusBot, TelegramBot
from alist_mikananirss.common.config_loader import ConfigLoader
from alist_mikananirss.core import (
    AnimeRenamer,
    DownloadManager,
    NotificationSender,
    RegexFilter,
    RssMonitor,
)
from alist_mikananirss.extractor import ChatGPTExtractor, Extractor

config_loader = None


def read_config(cfg_path: str):
    global config_loader
    config_loader = ConfigLoader(cfg_path)


def setup_logger():
    log_level = config_loader.get("dev.log_level", default="INFO")
    assert log_level in ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
    logger.remove()
    logger.add("log/main_{time}.log", retention="7 days", level=log_level)
    logger.add(sys.stderr, level=log_level)  # 添加新的 handler 且设置等级为 INFO


def setup_proxy():
    proxies = config_loader.get("common.proxies", {})
    if "http" in proxies:
        os.environ["HTTP_PROXY"] = proxies["http"]
    if "https" in proxies:
        os.environ["HTTPS_PROXY"] = proxies["https"]


async def init_alist():
    base_url = config_loader.get("alist.base_url")
    downloader_type = config_loader.get("alist.downloader")
    token = config_loader.get("alist.token")
    acfg = AlistConfig(base_url, token, downloader_type)
    alist_client = Alist(acfg)
    alist_ver = await alist_client.get_alist_ver()
    if alist_ver < "3.29.0":
        raise ValueError(f"Unsupported Alist version: {alist_ver}")
    return alist_client


def init_extrator():
    rename_cfg = config_loader.get("rename", None)
    if rename_cfg is None:
        return None
    if "chatgpt" in rename_cfg:
        chatgpt_cfg = rename_cfg["chatgpt"]
        chatgpt = ChatGPTExtractor(
            chatgpt_cfg["api_key"],
            chatgpt_cfg.get("base_url"),
            chatgpt_cfg.get("model"),
        )
        Extractor.initialize(chatgpt)
        return
    else:
        raise ValueError("Invalid rename config, extractor is required")


def init_notification_bots():
    notification_bots = []
    tg_config = config_loader.get("notification.telegram", None)
    if tg_config:
        bot_token = config_loader.get("notification.telegram.bot_token")
        user_id = config_loader.get("notification.telegram.user_id")
        bot = TelegramBot(bot_token, user_id)
        notification_bots.append(NotificationBot(bot))

    pushplus_config = config_loader.get("notification.pushplus", None)
    if pushplus_config:
        user_token = config_loader.get("notification.pushplus.token")
        channel = config_loader.get("notification.pushplus.channel", None)
        bot = PushPlusBot(user_token, channel)
        notification_bots.append(NotificationBot(bot))
    return notification_bots


def init_notification_sender():
    notification_bots = init_notification_bots()
    interval_time = config_loader.get("common.interval_time", 300)
    interval_time = config_loader.get("notification.interval_time", interval_time)
    NotificationSender.initialize(notification_bots, interval_time)


def init_resource_filter():
    regex_filter = RegexFilter()
    filters_name = config_loader.get("mikan.filters", [])
    regex_pattern = config_loader.get("mikan.regex_pattern", None)
    regex_filter.update_regex(regex_pattern)
    for name in filters_name:
        regex_filter.add_pattern(name)
    return regex_filter


def init_rss_monitor(regex_filter: RegexFilter):
    subscribe_url = config_loader.get("mikan.subscribe_url")
    use_extractor = False if config_loader.get("rename", None) is None else True
    rss_monitor = RssMonitor(
        subscribe_url,
        filter=regex_filter,
        use_extractor=use_extractor,
    )
    interval_time = config_loader.get("common.interval_time", 300)
    if interval_time < 0:
        raise ValueError("Invalid interval time")
    rss_monitor.set_interval_time(interval_time)
    return rss_monitor


def init_download_manager(alist_client: Alist):
    download_path = config_loader.get("alist.download_path")
    use_renamer = False if config_loader.get("rename", None) is None else True
    need_notification = (
        False if config_loader.get("notification", None) is None else True
    )
    DownloadManager.initialize(
        alist_client, download_path, use_renamer, need_notification
    )


def init_renamer(alist_client: Alist):
    def check_rename_format(rename_format: str):
        # Check if there are unknown keys in rename format
        from collections import defaultdict

        all_key_test_data = {
            "name": "test",
            "season": 1,
            "episode": 1,
            "fansub": "fansub",
            "quality": "1080p",
            "language": "简体中文",
            "ext": "mp4",
        }
        safe_dict = defaultdict(lambda: "undefined", all_key_test_data)
        res = rename_format.format_map(safe_dict)
        if "undefined" in res:
            unknown_keys = [
                key for key, value in safe_dict.items() if value == "undefined"
            ]
            raise KeyError(f"Error keys in rename format: {', '.join(unknown_keys)}")

    rename_cfg = config_loader.get("rename", None)
    if rename_cfg is not None:
        rename_format = config_loader.get("rename.rename_format", None)
        if rename_format:
            check_rename_format(rename_format)
        AnimeRenamer.initialize(alist_client, rename_format)



from .bot_base import BotBase  # noqa: F401
from .notificationbot import NotificationBot, NotificationMsg  # noqa: F401
from .tgbot import TelegramBot  # noqa: F401
from .pushplus_bot import PushPlusBot, PushPlusChannel  # noqa: F401



import asyncio
import os
import string

from loguru import logger

from alist_mikananirss.alist import Alist
from alist_mikananirss.websites import ResourceInfo


class AnimeRenamer:
    _instance = None
    _lock = asyncio.Lock()
    alist_client: Alist = None
    rename_format: str = "{name} S{season:02d}E{episode:02d}.{ext}"

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(AnimeRenamer, cls).__new__(cls)
        return cls._instance

    @classmethod
    def initialize(cls, alist: Alist, rename_format: str = None):
        instance = cls()
        instance.alist_client = alist
        if rename_format is not None:
            instance.rename_format = rename_format

    async def _build_new_name(self, old_filepath: str, resource: ResourceInfo):
        name = resource.anime_name
        season = resource.season
        episode = resource.episode
        if season is None or episode is None:
            raise ValueError("Season or episode is none when rename")
        fansub = resource.fansub
        quality = resource.quality
        language = resource.language
        old_filedir = os.path.dirname(old_filepath)
        old_filename = os.path.basename(old_filepath)
        file_ext = os.path.splitext(old_filename)[-1].replace(".", "")

        if season == 0:
            # 总集篇/OVA 则以顺序命名
            file_list = await self.alist_client.list_dir(old_filedir, per_page=999)
            episode = len(file_list)
        data = {
            "name": name,
            "season": season,
            "episode": episode,
            "ext": file_ext,
            "fansub": fansub,
            "quality": quality,
            "language": language,
        }
        data = {k: v for k, v in data.items() if v is not None}
        formatter = string.Formatter()
        result = []
        for literal_text, field_name, format_spec, _ in formatter.parse(
            self.rename_format
        ):
            if literal_text:
                result.append(literal_text)
            if field_name in data:
                value = data[field_name]
                value = format(value, format_spec)
                result.append(str(value))
        new_filename = "".join(result).strip()
        return new_filename

    @classmethod
    async def rename(cls, old_filepath: str, resource: ResourceInfo, max_retry=3):
        if (
            resource.anime_name is None
            or resource.season is None
            or resource.episode is None
        ):
            logger.error(f"rename failed due to resource info is invalid: {resource}")
            return
        instance = cls()
        for i in range(max_retry):
            try:
                new_filename = await instance._build_new_name(old_filepath, resource)
                await instance.alist_client.rename(old_filepath, new_filename)
                logger.info(f"Rename {old_filepath} to {new_filename}")
                break
            except Exception as e:
                if i < max_retry - 1:
                    logger.warning(f"Failed to rename {old_filepath}, retrying...: {e}")
                else:
                    logger.error(f"Error when rename {old_filepath}: {e}")
                await asyncio.sleep(5)



import os
from unittest.mock import MagicMock

import pytest
from alist_mikananirss.core import DownloadManager
from alist_mikananirss.websites.entities import ResourceInfo


@pytest.fixture
def base_path():
    return "/base/path"


def test_initialize(base_path):
    alist_mock = MagicMock()
    DownloadManager.initialize(alist_mock, base_path)
    assert DownloadManager.get_instance().base_download_path == base_path


def test_build_download_path_without_anime_name(base_path):
    resource = ResourceInfo(
        resource_title="Test Resource",
        torrent_url="http://example.com/torrent",
        published_date="2023-05-20",
    )
    expected_path = base_path
    test_instance = DownloadManager().get_instance()
    assert test_instance._build_download_path(resource) == expected_path


def test_build_download_path_with_anime_name(base_path):
    resource = ResourceInfo(
        resource_title="Test Resource",
        torrent_url="http://example.com/torrent",
        published_date="2023-05-20",
        anime_name="Test Anime",
    )
    expected_path = os.path.join(base_path, "Test Anime")
    test_instance = DownloadManager().get_instance()
    assert test_instance._build_download_path(resource) == expected_path


def test_build_download_path_with_anime_name_and_season(base_path):
    resource = ResourceInfo(
        resource_title="Test Resource",
        torrent_url="http://example.com/torrent",
        published_date="2023-05-20",
        anime_name="Test Anime",
        season=1,
    )
    expected_path = os.path.join(base_path, "Test Anime", "Season 1")
    test_instance = DownloadManager().get_instance()
    assert test_instance._build_download_path(resource) == expected_path


def test_build_download_path_with_season_no_anime_name(base_path):
    resource = ResourceInfo(
        resource_title="Test Resource",
        torrent_url="http://example.com/torrent",
        published_date="2023-05-20",
        season=1,
    )
    expected_path = base_path
    test_instance = DownloadManager().get_instance()
    assert test_instance._build_download_path(resource) == expected_path


def test_build_download_path_with_special_characters(base_path):
    resource = ResourceInfo(
        resource_title="Test Resource",
        torrent_url="http://example.com/torrent",
        published_date="2023-05-20",
        anime_name="Test:Anime!",
        season=2,
    )
    expected_path = os.path.join(base_path, "Test Anime!", "Season 2")
    test_instance = DownloadManager().get_instance()
    assert test_instance._build_download_path(resource) == expected_path


def test_build_download_path_with_empty_anime_name(base_path):
    resource = ResourceInfo(
        resource_title="Test Resource",
        torrent_url="http://example.com/torrent",
        published_date="2023-05-20",
        anime_name="",
        season=3,
    )
    expected_path = base_path
    test_instance = DownloadManager().get_instance()
    assert test_instance._build_download_path(resource) == expected_path


def test_build_download_path_with_none_anime_name(base_path):
    resource = ResourceInfo(
        resource_title="Test Resource",
        torrent_url="http://example.com/torrent",
        published_date="2023-05-20",
        anime_name=None,
        season=4,
    )
    expected_path = base_path
    test_instance = DownloadManager().get_instance()
    assert test_instance._build_download_path(resource) == expected_path


def test_build_download_path_with_special_season(base_path):
    resource = ResourceInfo(
        resource_title="Test Resource",
        torrent_url="http://example.com/torrent",
        published_date="2023-05-20",
        anime_name="Test Anime",
        season=0,
    )
    expected_path = os.path.join(base_path, "Test Anime", "Season 0")
    test_instance = DownloadManager().get_instance()
    assert test_instance._build_download_path(resource) == expected_path



from dataclasses import dataclass
from typing import Optional


@dataclass
class ResourceInfo:
    resource_title: str
    torrent_url: str

    published_date: Optional[str] = None
    anime_name: Optional[str] = None
    season: Optional[int] = None
    episode: Optional[int] = None
    fansub: Optional[str] = None
    quality: Optional[str] = None
    language: Optional[str] = None

    def __hash__(self):
        return hash(self.resource_title)


@dataclass
class FeedEntry:
    resource_title: str
    torrent_url: str
    published_date: Optional[str] = None
    homepage_url: Optional[str] = None

    def __hash__(self):
        return hash(self.resource_title)



from unittest.mock import AsyncMock, patch

import pytest
from alist_mikananirss.extractor import (
    AnimeNameExtractResult,
    ChatGPTExtractor,
    ResourceTitleExtractResult,
)


@pytest.fixture
def chatgpt_extractor():
    with patch("openai.AsyncOpenAI", new=AsyncMock()):
        yield ChatGPTExtractor("fake_api_key")


@pytest.mark.asyncio
async def test_analyse_anime_name_success(chatgpt_extractor):
    mock_response = '```json\n{"anime_name": "Test Anime", "season": 2}\n```'

    with patch.object(
        chatgpt_extractor, "_get_gpt_response", new_callable=AsyncMock
    ) as mock_get_response:
        mock_get_response.return_value = mock_response

        result = await chatgpt_extractor.analyse_anime_name("Test Anime Season 2")

        assert isinstance(result, AnimeNameExtractResult)
        assert result.anime_name == "Test Anime"
        assert result.season == 2


@pytest.mark.asyncio
async def test_analyse_anime_name_invalid_json(chatgpt_extractor):
    mock_response = "Invalid JSON response"

    with patch.object(
        chatgpt_extractor, "_get_gpt_response", new_callable=AsyncMock
    ) as mock_get_response:
        mock_get_response.return_value = mock_response

        with pytest.raises(ValueError, match="Can't parse GPT responese as a json"):
            await chatgpt_extractor.analyse_anime_name("Test Anime Season 2")


@pytest.mark.asyncio
async def test_analyse_anime_name_wrong_type(chatgpt_extractor):
    mock_response = '```json\n{"anime_name": "Test Anime", "season": "4"}\n```'

    with patch.object(
        chatgpt_extractor, "_get_gpt_response", new_callable=AsyncMock
    ) as mock_get_response:
        mock_get_response.return_value = mock_response

        with pytest.raises(TypeError, match="GPT provide a wrong type data"):
            await chatgpt_extractor.analyse_anime_name("Test Anime Season 2")


@pytest.mark.asyncio
async def test_analyse_resource_title_success(chatgpt_extractor):
    mock_response = """
    ```json
    {
        "anime_name_cn": "测试番剧",
        "anime_name_jp": "测试番剧",
        "anime_name_en": "Test Anime",
        "season": 2,
        "episode": 1,
        "quality": "1080p",
        "fansub": "kuonji alice My wife",
        "language": "简体中文"
    }
    ```
    """

    with patch.object(
        chatgpt_extractor, "_get_gpt_response", new_callable=AsyncMock
    ) as mock_get_response:
        mock_get_response.return_value = mock_response

        result = await chatgpt_extractor.analyse_resource_title(
            "[kuonji alice My wife] 测试番剧 第二季 / Test Anime S2 - 01 [1080p][简体]"
        )

        assert isinstance(result, ResourceTitleExtractResult)
        assert result.anime_name == "测试番剧"
        assert result.season == 2
        assert isinstance(result.episode, int)
        assert result.episode == 1
        assert result.quality == "1080p"
        assert result.fansub == "kuonji alice My wife"
        assert result.language == "简体中文"


@pytest.mark.asyncio
async def test_analyse_resource_title_special_episode(chatgpt_extractor):
    mock_response = """```json
    {
        "anime_name_cn": "测试番剧",
        "anime_name_jp": "测试番剧",
        "anime_name_en": "Test Anime",
        "season": 0,
        "episode": 1.5,
        "quality": "720p",
        "fansub": "kuonji alice My wife",
        "language": "简体中文"
    }
    ```"""

    with patch.object(
        chatgpt_extractor, "_get_gpt_response", new_callable=AsyncMock
    ) as mock_get_response:
        mock_get_response.return_value = mock_response

        result = await chatgpt_extractor.analyse_resource_title(
            "[kuonji alice My wife] 测试番剧 第二季 / Test Anime S2 - 1.5 [720p][简体]"
        )

        assert isinstance(result, ResourceTitleExtractResult)
        assert result.anime_name == "测试番剧"
        assert result.season == 0
        assert isinstance(result.episode, int)
        assert result.episode == 0
        assert result.quality == "720p"
        assert result.fansub == "kuonji alice My wife"
        assert result.language == "简体中文"


@pytest.mark.asyncio
async def test_analyse_resource_title_invalid_json(chatgpt_extractor):
    mock_response = "Invalid JSON response"

    with patch.object(
        chatgpt_extractor, "_get_gpt_response", new_callable=AsyncMock
    ) as mock_get_response:
        mock_get_response.return_value = mock_response

        with pytest.raises(ValueError, match="Can't parse GPT responese as a json"):
            await chatgpt_extractor.analyse_resource_title(
                "[kuonji alice My wife] 测试番剧 第二季 / Test Anime S2 - 01 [1080p][简体]"
            )


@pytest.mark.asyncio
async def test_analyse_resource_title_wrong_type(chatgpt_extractor):
    mock_response = """```json
    {
        "anime_name_cn": "测试番剧",
        "anime_name_jp": "测试番剧",
        "anime_name_en": "Test Anime",
        "season": "2",
        "episode": 1,
        "quality": "1080p",
        "fansub": "kuonji alice My wife",
        "language": "简体中文"
    }
    ```"""

    with patch.object(
        chatgpt_extractor, "_get_gpt_response", new_callable=AsyncMock
    ) as mock_get_response:
        mock_get_response.return_value = mock_response

        with pytest.raises(TypeError, match="GPT provide a wrong type data"):
            await chatgpt_extractor.analyse_resource_title(
                "[kuonji alice My wife] 测试番剧 第二季 / Test Anime S2 - 01 [1080p][简体]"
            )



import abc
import asyncio
from concurrent.futures import ThreadPoolExecutor

import feedparser

from .entities import FeedEntry, ResourceInfo


class Website(abc.ABC):
    """Website，虚基类，提供从各站点Rss链接中提取资源信息的接口"""

    def __init__(self, rss_url: str):
        self.rss_url = rss_url

    async def parse_feed(self, url):
        """使用feedparser库异步解析rss链接"""
        loop = asyncio.get_event_loop()
        with ThreadPoolExecutor() as pool:
            feed = await loop.run_in_executor(pool, feedparser.parse, url)
        return feed

    @abc.abstractmethod
    async def get_feed_entries(self, rss_url: str) -> set[FeedEntry]:
        """从rss链接中获取所有的资源条目"""
        pass

    @abc.abstractmethod
    async def extract_resource_info(
        self, entry: FeedEntry, use_extractor: bool = False
    ) -> ResourceInfo:
        """从资源条目中提取番剧资源信息"""
        pass


class WebsiteFactory:
    """Website工厂类，根据rss链接创建对应的Website类"""

    @staticmethod
    def get_website_parser(rss_url: str) -> Website:
        if "mikan" in rss_url:
            from alist_mikananirss.websites import Mikan

            return Mikan(rss_url)
        else:
            raise ValueError("Unknown website")



from .download_manager import DownloadManager
from .filters import RegexFilter
from .notification_sender import NotificationSender
from .renamer import AnimeRenamer
from .rss_monitor import RssMonitor



import asyncio
import os
import time
from dataclasses import dataclass
from typing import Optional

from loguru import logger

from alist_mikananirss.alist import Alist
from alist_mikananirss.alist.tasks import (
    AlistDownloadTask,
    AlistTaskCollection,
    AlistTaskStatus,
    AlistTaskType,
    AlistTransferTask,
)
from alist_mikananirss.common.database import SubscribeDatabase
from alist_mikananirss.websites import ResourceInfo

from .notification_sender import NotificationSender
from .renamer import AnimeRenamer


@dataclass
class AnimeDownloadTaskInfo:
    resource: ResourceInfo
    download_path: str
    download_task: AlistDownloadTask
    transfer_task: Optional[AlistTransferTask] = None


class TaskMonitor:
    normal_status = [
        AlistTaskStatus.Pending,
        AlistTaskStatus.Running,
        AlistTaskStatus.StateBeforeRetry,
        AlistTaskStatus.StateWaitingRetry,
    ]

    def __init__(
        self,
        alist: Alist,
        task: AlistTransferTask | AlistDownloadTask,
    ):
        self.alist = alist
        self.task = task

    async def __refresh(self):
        task_list = await self.alist.get_task_list(self.task.task_type)
        task = task_list[self.task.tid]
        if not task:
            raise RuntimeError(f"Can't find the task {self.task.tid}")
        self.task = task

    async def wait_finished(self):
        last_progress = 0
        last_progress_time = time.time()
        stall_threshold = 300  # 5分钟
        progress_threshold = 0.01  # 1%
        while True:
            try:
                await self.__refresh()
                current_progress = self.task.progress
            except Exception as e:
                logger.warning(f"Error when refresh {self.task} status: {e}")
                await asyncio.sleep(1)
                continue
            logger.debug(
                f"Checking {self.task} status: {self.task.status} progress: {current_progress:.2f}%"
            )
            if self.task.status not in self.normal_status:
                return self.task

            # 若长时间下载进度没有变化，则抛出异常
            current_time = time.time()
            time_elapsed = current_time - last_progress_time
            progress_change = current_progress - last_progress
            if time_elapsed > stall_threshold and progress_change < progress_threshold:
                raise RuntimeError(
                    f"Progress is too slow: {progress_change:.2%} in {time_elapsed:.2f} seconds"
                )
            if progress_change >= progress_threshold:
                last_progress = current_progress
                last_progress_time = current_time

            await asyncio.sleep(1)


class DownloadManager:
    _instance = None
    alist_client: Alist
    base_download_path: str
    use_renamer: bool = False
    need_notification: bool = False
    uuid_set: set[str] = set()
    db = SubscribeDatabase()

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(DownloadManager, cls).__new__(cls)
        return cls._instance

    @classmethod
    def initialize(
        cls,
        alist_client: Alist,
        base_download_path: str,
        use_renamer: bool = False,
        need_notification: bool = False,
    ) -> None:
        cls._instance = cls()
        cls.alist_client = alist_client
        cls.base_download_path = base_download_path
        cls.use_renamer = use_renamer
        cls.need_notification = need_notification

    @classmethod
    def get_instance(cls):
        if not cls._instance:
            raise RuntimeError("DownloadManager is not initialized")
        return cls._instance

    @classmethod
    async def add_download_tasks(cls, resources: list[ResourceInfo]):
        instance = cls.get_instance()
        info_list: list[AnimeDownloadTaskInfo] = await instance.download(resources)
        for task_info in info_list:
            instance.db.insert_resource_info(task_info.resource)
            asyncio.create_task(instance.monitor(task_info))

    async def _find_transfer_task(self, resource: ResourceInfo) -> AlistTransferTask:
        def is_video(file_name: str) -> bool:
            video_suffix = [".mp4", ".mkv", ".avi", ".rmvb", ".wmv", ".flv"]
            for suffix in video_suffix:
                if file_name.endswith(suffix):
                    return True
            return False

        async with asyncio.timeout(10):
            while True:
                try:
                    transfer_task_list = await self.alist_client.get_task_list(
                        AlistTaskType.TRANSFER
                    )
                    for transfer_task in transfer_task_list:
                        # 查找第一个，未被标记过的番剧名相同的视频文件传输任务作为下载任务对应的传输任务
                        if (
                            transfer_task.uuid not in self.uuid_set
                            and is_video(transfer_task.file_name)
                            and transfer_task.status
                            in [AlistTaskStatus.Pending, AlistTaskStatus.Running]
                            and resource.anime_name in transfer_task.description
                        ):

                            self.uuid_set.add(transfer_task.uuid)
                            logger.debug(
                                f"Linked {resource.resource_title} to {transfer_task.uuid}"
                            )
                            return transfer_task

                    logger.warning(
                        f"Can't find the transfer task of {resource.resource_title}"
                    )
                    await asyncio.sleep(1)
                except Exception as e:
                    logger.error(f"Error when getting transfer task list: {e}")
                    await asyncio.sleep(1)

    async def _wait_finished(
        self, task: AnimeDownloadTaskInfo
    ) -> Optional[AnimeDownloadTaskInfo]:
        """monitor one task until it succeed

        Args:
            resource (MikanAnimeResource): resource need to be monitored

        Returns:
            MikanAnimeResource | None: return the resource if download succeed, else return None
        """
        download_task = task.download_task
        download_task_monitor = TaskMonitor(self.alist_client, download_task)
        try:
            download_task = await download_task_monitor.wait_finished()
        except RuntimeError:
            self.alist_client.cancel_task(download_task)
            logger.error(
                f"Timeout to wait the download task of {task.resource.resource_title} succeed"
            )
            return None
        if download_task.status != AlistTaskStatus.Succeeded:
            logger.error(f"Error when download {task.resource.resource_title}")
            return None

        try:
            transfer_task = await self._find_transfer_task(task.resource)
        except asyncio.TimeoutError:
            logger.error(
                f"Timeout to find the transfer task of {task.resource.resource_title}"
            )
            return None
        transfer_task_monitor = TaskMonitor(self.alist_client, transfer_task)
        try:
            transfer_task = await transfer_task_monitor.wait_finished()
        except RuntimeError:
            self.alist_client.cancel_task(transfer_task)
            logger.error(
                f"Timeout to wait the transfer task of {task.resource.resource_title} succeed"
            )
            return None
        if transfer_task.status != AlistTaskStatus.Succeeded:
            logger.error(f"Error when transfer {task.resource.resource_title}")
            return None
        task.transfer_task = transfer_task
        return task

    def _post_process(self, task: AnimeDownloadTaskInfo):
        if self.use_renamer:
            filepath = os.path.join(task.download_path, task.transfer_task.file_name)
            asyncio.create_task(AnimeRenamer.rename(filepath, task.resource))
        if self.need_notification:
            asyncio.create_task(NotificationSender.add_resource(task.resource))

    async def monitor(self, task_info: AnimeDownloadTaskInfo):
        success_task = await self._wait_finished(task_info)
        if success_task is not None:
            self._post_process(success_task)
        else:
            # 下载失败，删除数据库记录
            self.db.delete_by_resource_title(task_info.resource.resource_title)

    def _build_download_path(self, resource: ResourceInfo) -> str:
        """build the download path based on the anime name and season
        The result looks like this: base_download_path/AnimeName/Season {season} \n
        if the anime name is not available, use the base download path

        Args:
            resource (ResourceInfo)

        Returns:
            str: new download path
        """
        download_path = self.base_download_path
        if resource.anime_name:
            # Replace illegal characters in the anime_name
            illegal_chars = ["\\", "/", ":", "*", "?", '"', "<", ">", "|"]
            anime_name = resource.anime_name
            for char in illegal_chars:
                anime_name = anime_name.replace(char, " ")
            download_path = os.path.join(download_path, anime_name)
            # Need to compare with None, or when seaon=0 will be ignored
            if resource.season is not None:
                download_path = os.path.join(download_path, f"Season {resource.season}")
        return download_path

    async def download(
        self, new_resources: list[ResourceInfo]
    ) -> list[AnimeDownloadTaskInfo]:
        """Create alist offline download task

        Args:
            new_resources (list[MikanAnimeResource]): resources list
            base_download_path (str): remote dir path

        Returns:
            list[AnimeDownloadTaskInfo]: Successful download task's info
        """
        # Generate a mapping of download paths to resources
        # facilitating batch creation of download tasks and reducing requests to the Alist API
        ## mapping of {download_path: [torrent_url]}
        path_urls: dict[str, list[str]] = {}
        resource_path_map = {}  # mapping of {ResourceInfo: download_path}
        for resource in new_resources:
            download_path = self._build_download_path(resource)
            path_urls.setdefault(download_path, []).append(resource.torrent_url)
            resource_path_map[resource] = download_path
        # start to request the Alist Download API
        task_list = AlistTaskCollection()
        for download_path, urls in path_urls.items():
            try:
                tmp_task_list = await self.alist_client.add_offline_download_task(
                    download_path, urls
                )
            except Exception as e:
                logger.error(f"Error when add offline download task: {e}")
                continue
            task_list = task_list + tmp_task_list
        # Patch the download task with the resource information
        anime_task_list = []
        for task in task_list:
            for resource in new_resources:
                if resource.torrent_url == task.url:
                    anime_task = AnimeDownloadTaskInfo(
                        resource=resource,
                        download_task=task,
                        download_path=resource_path_map[resource],
                    )
                    logger.info(
                        f"Start to download {resource.resource_title} to {anime_task.download_path}"
                    )
                    anime_task_list.append(anime_task)
                    break
        return anime_task_list



import yaml


class ConfigLoader:
    _MISSING = object()

    def __init__(self, config_path):
        self.config_path = config_path
        self.config = self.load_config()

    def load_config(self):
        with open(self.config_path, "r", encoding="utf-8") as file:
            return yaml.safe_load(file)

    def get(self, path, default=_MISSING):
        keys = path.split(".")
        value = self.config
        for key in keys:
            if value.get(key) is None:
                if default is not self._MISSING:
                    return default
                else:
                    raise KeyError(
                        f"{path} is not found in config file {self.config_path}"
                    )
            value = value[key]
        return value



import pytest
import yaml
from alist_mikananirss.common.config_loader import ConfigLoader


@pytest.fixture
def sample_config():
    return {
        "common": {
            "interval_time": 300,
            "proxies": {
                "http": "http://127.0.0.1:7890",
                "https": "http://127.0.0.1:7890",
            },
        },
        "alist": {
            "base_url": "http://www.example.com",
            "token": "alist-xxx",
            "downloader": "aria2",
            "download_path": "Onedrive/Anime",
        },
        "mikan": {
            "subscribe_url": [
                "https://mikanani.me/RSS/MyBangumi?token=xxx",
                "https://mikanani.me/RSS/rss2",
            ],
            "filters": ["1080p", "非合集"],
        },
    }


@pytest.fixture
def config_file(tmp_path, sample_config):
    config_file = tmp_path / "test_config.yaml"
    with open(config_file, "w") as f:
        yaml.dump(sample_config, f)
    return config_file


def test_load_config(config_file, sample_config):
    loader = ConfigLoader(str(config_file))
    assert loader.config == sample_config


def test_get_existing_key(config_file):
    loader = ConfigLoader(str(config_file))
    assert loader.get("common.interval_time") == 300
    assert loader.get("alist.base_url") == "http://www.example.com"
    assert loader.get("mikan.subscribe_url") == [
        "https://mikanani.me/RSS/MyBangumi?token=xxx",
        "https://mikanani.me/RSS/rss2",
    ]


def test_get_nested_key(config_file):
    loader = ConfigLoader(str(config_file))
    assert loader.get("common.proxies.http") == "http://127.0.0.1:7890"


def test_get_non_existent_key_with_default(config_file):
    loader = ConfigLoader(str(config_file))
    assert loader.get("non_existent_key", default="default_value") == "default_value"


def test_get_non_existent_key_without_default(config_file):
    loader = ConfigLoader(str(config_file))
    with pytest.raises(KeyError):
        loader.get("non_existent_key")


def test_get_partial_non_existent_key(config_file):
    loader = ConfigLoader(str(config_file))
    with pytest.raises(KeyError):
        loader.get("common.non_existent_key")



from dataclasses import dataclass
from typing import Optional


@dataclass
class ResourceTitleExtractResult:
    anime_name: str
    season: int
    episode: int
    quality: Optional[str] = None
    fansub: Optional[str] = None
    language: Optional[str] = None


@dataclass
class AnimeNameExtractResult:
    anime_name: str
    season: int



import os
import sqlite3
from datetime import datetime

from loguru import logger

from alist_mikananirss.websites import ResourceInfo

db_path = "data"
os.makedirs(db_path, exist_ok=True)


class SubscribeDatabase:
    def __init__(self, db_name="subscribe_database.db"):
        self.db_name = os.path.join(db_path, db_name)
        self.conn = None
        self.cursor = None
        self.__check()

    def connect(self):
        self.conn = sqlite3.connect(self.db_name)
        self.cursor = self.conn.cursor()

    def close(self):
        if self.conn:
            self.conn.close()

    def __check(self):
        if not os.path.exists(self.db_name):
            self.__create_table()
        else:
            self._upgrade_database()

    def __create_table(self):
        self.connect()
        self.cursor.execute(
            """
            CREATE TABLE IF NOT EXISTS resource_data (
                id TEXT PRIMARY KEY,
                title TEXT,
                link TEXT,
                published_date TEXT,
                downloaded_date TEXT,
                anime_name TEXT,
                season INTEGER,
                episode INTEGER,
                fansub TEXT,
                quality TEXT,
                language TEXT
            )
            """
        )
        self.cursor.execute(
            """
            CREATE TABLE IF NOT EXISTS db_version (
                version INTEGER PRIMARY KEY
            )
            """
        )
        self.cursor.execute("INSERT INTO db_version (version) VALUES (1)")
        self.conn.commit()
        self.close()

    def _upgrade_database(self):
        self.connect()
        try:
            self.cursor.execute("SELECT version FROM db_version")
            version = self.cursor.fetchone()[0]
        except sqlite3.OperationalError:
            self.cursor.execute(
                """
                CREATE TABLE IF NOT EXISTS db_version (
                    version INTEGER PRIMARY KEY
                )
            """
            )
            version = 0

        if version < 1:
            try:
                self.cursor.execute(
                    """
                    CREATE TABLE resource_data_new (
                        id INTEGER PRIMARY KEY AUTOINCREMENT,
                        resource_title TEXT NOT NULL,
                        torrent_url TEXT UNIQUE,
                        published_date TEXT,
                        downloaded_date TEXT,
                        anime_name TEXT,
                        season INTEGER,
                        episode INTEGER,
                        fansub TEXT,
                        quality TEXT,
                        language TEXT
                    )
                """
                )
                # 迁移数据
                self.cursor.execute(
                    """
                    INSERT INTO resource_data_new (resource_title, torrent_url, published_date, downloaded_date, anime_name)
                    SELECT title, link, published_date, downloaded_date, anime_name FROM resource_data
                """
                )
                # 删除旧表
                self.cursor.execute("DROP TABLE resource_data")
                # 重命名新表
                self.cursor.execute(
                    "ALTER TABLE resource_data_new RENAME TO resource_data"
                )
                # 创建新的索引
                self.cursor.execute(
                    """
                    CREATE UNIQUE INDEX idx_title
                    ON resource_data(resource_title)
                """
                )
                self.cursor.execute(
                    "INSERT OR REPLACE INTO db_version (version) VALUES (1)"
                )
                self.conn.commit()
                logger.info("Database upgraded to version 1")
            except Exception as e:
                logger.error(f"Error during database upgrade: {e}")
                self.conn.rollback()
        self.close()

    def insert(
        self,
        resource_title,
        torrent_url,
        published_date,
        downloaded_date,
        anime_name,
        season=None,
        episode=None,
        fansub=None,
        quality=None,
        language=None,
    ):
        self.connect()
        try:
            self.cursor.execute(
                """
                INSERT INTO resource_data 
                (resource_title, torrent_url, published_date, downloaded_date, anime_name, season, episode, fansub, quality, language)
                VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?)
            """,
                (
                    resource_title,
                    torrent_url,
                    published_date,
                    downloaded_date,
                    anime_name,
                    season,
                    episode,
                    fansub,
                    quality,
                    language,
                ),
            )
            self.conn.commit()
            logger.debug(f"Insert new resource: {anime_name}, {resource_title}")
        except sqlite3.IntegrityError:
            logger.debug(f"Resource already exists: {anime_name}, {resource_title}")
        except Exception as e:
            logger.error(f"Error when inserting resource: {e}")
        finally:
            self.close()

    def insert_resource_info(self, resource: ResourceInfo):
        downloaded_date = datetime.now().strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3]
        self.insert(
            resource.resource_title,
            resource.torrent_url,
            resource.published_date,
            downloaded_date,
            resource.anime_name,
            season=resource.season,
            episode=resource.episode,
            fansub=resource.fansub,
            quality=resource.quality,
            language=resource.language,
        )

    def is_resource_title_exist(self, resource_title: str):
        self.connect()
        try:
            self.cursor.execute(
                "SELECT 1 FROM resource_data WHERE resource_title = ? LIMIT 1",
                (resource_title,),
            )
            return self.cursor.fetchone() is not None
        except Exception as e:
            logger.error(f"Error checking resource existence: {e}")
            return False
        finally:
            self.close()

    def delete_by_id(self, id):
        self.connect()
        try:
            self.cursor.execute("DELETE FROM resource_data WHERE id=?", (id,))
            self.conn.commit()
            logger.debug(f"Delete resource data: {id}")
        except Exception as e:
            logger.error(f"Error when delete resource data:\n {e}")
        finally:
            self.close()

    def delete_by_torrent_url(self, url: str):
        self.connect()
        try:
            self.cursor.execute("DELETE FROM resource_data WHERE torrent_url=?", (url,))
            self.conn.commit()
            logger.debug(f"Delete resource data: {url}")
        except Exception as e:
            logger.error(f"Error when delete resource data:\n {e}")
        finally:
            self.close()

    def delete_by_resource_title(self, title: str):
        self.connect()
        try:
            self.cursor.execute(
                "DELETE FROM resource_data WHERE resource_title=?", (title,)
            )
            self.conn.commit()
            logger.debug(f"Delete resource data: {title}")
        except Exception as e:
            logger.error(f"Error when delete resource data:\n {e}")
        finally:
            self.close()



import asyncio
from typing import List

from loguru import logger

from alist_mikananirss.bot import NotificationBot, NotificationMsg
from alist_mikananirss.websites import ResourceInfo


class NotificationSender:
    _instance = None
    notification_bots: List[NotificationBot] = []
    _queue: asyncio.Queue = None
    _interval: int = 60

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    @classmethod
    def get_instance(cls):
        if cls._instance is None:
            raise RuntimeError("NotificationSender is not initialized")
        return cls._instance

    @classmethod
    def initialize(cls, notification_bots: List[NotificationBot], interval: int = 60):
        cls._instance = NotificationSender()
        cls._instance.notification_bots = notification_bots
        cls._instance._queue = asyncio.Queue()
        cls._instance._interval = interval
        asyncio.create_task(cls._instance._run())

    @classmethod
    def set_notification_bots(cls, notification_bots: List[NotificationBot]):
        cls._instance.notification_bots = notification_bots

    @classmethod
    async def add_resource(cls, resource: ResourceInfo):
        await cls._instance._queue.put(resource)

    @classmethod
    def set_interval(cls, interval: int):
        cls._instance._interval = interval

    async def _run(self):
        while True:
            await asyncio.sleep(self._interval)
            resources = []
            while not self._queue.empty():
                try:
                    resource = self._queue.get_nowait()
                    resources.append(resource)
                except asyncio.QueueEmpty:
                    break
            if resources:
                await self._send(resources)

    async def _send(self, resources: List[ResourceInfo]):
        if not self.notification_bots:
            return
        msg = NotificationMsg.from_resources(resources)
        logger.debug(f"Send notification\n: {msg}")
        results = await asyncio.gather(
            *[bot.send_message(msg) for bot in self.notification_bots],
            return_exceptions=True,
        )
        for result in results:
            if isinstance(result, Exception):
                logger.error(result)



import json
import re

from loguru import logger
from openai import AsyncOpenAI

from .base import ExtractorBase
from .models import AnimeNameExtractResult, ResourceTitleExtractResult


class ChatGPTExtractor(ExtractorBase):
    def __init__(self, api_key, base_url=None, model="gpt-3.5-turbo") -> None:
        self._api_key = api_key
        self._base_url = base_url
        self.model = model
        self._client = None

    @property
    def client(self):
        if self._client is None:
            self._client = AsyncOpenAI(api_key=self._api_key)
            if self._base_url:
                self._client.base_url = self._base_url
        return self._client

    async def _get_gpt_response(self, prompt, resource_name):
        chat_completion = await self.client.chat.completions.create(
            messages=[
                {"role": "system", "content": prompt},
                {"role": "user", "content": resource_name},
            ],
            model=self.model,
            temperature=0,
        )
        return chat_completion.choices[0].message.content

    async def _parse_json_response(self, resp):
        match = re.search(r"(\{.*?\})", resp, re.DOTALL)
        if not match:
            raise ValueError(f"Can't parse GPT responese as a json:\n {resp}")
        return json.loads(match.group(1))

    async def analyse_anime_name(self, anime_name: str) -> AnimeNameExtractResult:
        prompt = """
        In the following, I will provide you with an anime name. Please extract the original name of the anime (i.e., the name without season information) and the season number based on the given name.
        I need to parse this text into a data structure to initialize a AnimeNameInfo class in my code. The definition of the AnimeNameInfo class is as follows:

        class AnimeNameInfo{
            string anime_name;
            int season;
            
            MikanAnimeResource(this.anime_name, this.season);
        }

        Based on the resource name of the anime series, please provide a JSON-formatted data structure(output in markdown format) that I can directly use to initialize an instance of the class.
        """
        resp = await self._get_gpt_response(prompt, anime_name)
        data = await self._parse_json_response(resp)

        expected_data = {"anime_name": str, "season": int}
        if not all(
            isinstance(data.get(field_name), field_type)
            for field_name, field_type in expected_data.items()
        ):
            raise TypeError(f"GPT provide a wrong type data: {data}")

        info = AnimeNameExtractResult(
            anime_name=data["anime_name"], season=data["season"]
        )
        logger.debug(f"Chatgpt analyse resource name: {anime_name} -> {info}")
        return info

    async def analyse_resource_title(
        self, resource_title: str
    ) -> ResourceTitleExtractResult:
        prompt = """
        I will provide you with the torrent name of an anime. Please extract the following information from the torrent name: the name of the anime; the season number of the anime(If this episode is OVA，season set to 0); the episode number of the anime; the video quality; the fansub's name; and the language of the subtitles. 
        I need to parse this text into a data structure to initialize a ResourceNameInfo class in my code. The definition of the ResourceNameInfo class is as follows:

        class ResourceNameInfo{
            string anime_name_cn;
            string anime_name_jp;
            string anime_name_en;
            int season;
            float episode;
            string quality;
            string fansub;
            string language;
            
            ResourceNameInfo(this.anime_name, this.season, this.episode, this.quality, this.fansub, this.language);
        }

        Based on the anime resource name, please provide a JSON format data structure(output in markdown format) that I can use directly to initialize an instance of the ResourceNameInfo class.
        ps:
        1. If the episode number is a decimal, it means that it is a special episode. In this case, the season number should be set to 0.
        2. Assume season=1 if no special indication is given
        3. please make a comprehensive judgment based on the values of anime_name_cn, anime_name_jp, and anime_name_en. The meanings of these three should be relatively similar.
        4. If there are multiple names, please choose just one of them
        5. anime name should not contains season info
        """
        resp = await self._get_gpt_response(prompt, resource_title)
        data = await self._parse_json_response(resp)

        expected_data = {
            "anime_name_cn": str,
            "anime_name_jp": str,
            "anime_name_en": str,
            "season": int,
            "episode": (float, int),
            "quality": str,
            "fansub": str,
            "language": str,
        }

        if not all(
            isinstance(data.get(field_name), field_type)
            for field_name, field_type in expected_data.items()
        ):
            raise TypeError(f"GPT provide a wrong type data: {data}")
        info = ResourceTitleExtractResult(
            anime_name=data["anime_name_cn"],
            season=data["season"],
            episode=int(data["episode"]) if data["season"] != 0 else 0,
            quality=data["quality"],
            fansub=data["fansub"],
            language=data["language"],
        )
        logger.debug(f"Chatgpt analyse resource name: {resource_title} -> {info}")
        return info
